#!/usr/bin/python
#
# Copyright (C) 2010-2011, Canonical Ltd (http://www.canonical.com/)
#
# This file is part of ubuntu-iso-testing.
#
# ubuntu-iso-testing is free software: you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# ubuntu-iso-testing is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ubuntu-iso-testing.  If not, see
# <http://www.gnu.org/licenses/>.
#
from lockfile import FileLock
import os
import optparse
import sys
import logging
import urllib2
import subprocess
import hashlib
from time import sleep

# Default options
DEFAULT_VARIANT = 'server'
try:
    import distro_info
    DEFAULT_RELEASE = distro_info.UbuntuDistroInfo().devel()
except ImportError:
    DEFAULT_RELEASE = 'trusty'

DEFAULT_ARCH = 'all'
DEFAULT_FLAVOR = 'ubuntu-server'
DEFAULT_ISOROOT = os.path.expanduser('~/isos')

TIMEOUT = 15 # Timeout between retries when the image is not on the server
RETRY_MAX = 4 # Number of retries

# URLS
BASE_URL = "http://cdimage.ubuntu.com"

# List of 'all' architectures
ALL_ARCHS = [ 'i386',  'amd64']

class HeadRequest(urllib2.Request):
    """ Used to check to retrieve headers of a remote file """
    def get_method(self):
        return "HEAD"

# Calculates the sha256 hash for a given file
def validate_sha256(fname, sha256_hash, block_size=2**20):
    """ Validate checksum of file fname against sha256 hash """
    logging.debug("Calculating hash for file '%s'", fname)
    sha256 = hashlib.sha256()
    with open(fname) as f:
        while True:
            data = f.read(block_size)
            if not data:
                break
            sha256.update(data)

    l_sha256 = sha256.hexdigest()
    logging.debug("Expected Checksum: '%s'", sha256_hash)
    logging.debug("Local File Checksum: '%s'", l_sha256)
    return sha256_hash == l_sha256

def download_iso(isoroot, release, variant, arch, flavor=None, no_act=False):
    """ Download an iso """
    logging.info("Downloading ISO...")
    cmd = ['dl-ubuntu-test-iso', '--variant=%s' % variant,
                    '--arch=%s' % arch, '--release=%s' % release,
                    '--build=current' ,'--isoroot=%s' % isoroot, '-P']

    if flavor:
        cmd.append('--flavor=%s' % flavor)

    if no_act:
        cmd.append('-n')

    logging.debug("Cmd: %s", cmd)
 
    try:
        subprocess.check_call(cmd)
    except subprocess.CalledProcessError, e:
        logging.error(e)
        return e.returncode
    return 0

def remote_file_exists(file_url):
    """ Check if the remote file exists

    Return file info on success false otherwise """
    logging.debug("Checking if remote file exists: '%s'", file_url)
    try:
        res = urllib2.urlopen(HeadRequest(file_url))
        logging.debug('URL Found:\n%s', res.info())
        return res.info()
    except (urllib2.HTTPError, urllib2.URLError), e:
        logging.debug("Check failed: %s", e)
        return False

def main():
    """ Main routine """
    usage = "Usage: %prog [options]"
    parser = optparse.OptionParser(usage=usage)
    parser.add_option("-d", "--debug", dest="debug", action="store_true",
                      default=False, help="enable debugging")
    parser.add_option("-r", "--release",  dest="release",
                      default=DEFAULT_RELEASE,
                      help="release of Ubuntu to download (default=%s)"
                      % DEFAULT_RELEASE)
    parser.add_option("-v", "--variant",  dest="variant",
                      default=DEFAULT_VARIANT,
                      help="variant of Ubuntu to download (default=%s)"
                      % DEFAULT_VARIANT)
    parser.add_option("-a", "--arch",  dest="arch", default=DEFAULT_ARCH,
                      help="arch of Ubuntu to download (default=%s)"
                      % DEFAULT_ARCH)
    parser.add_option("-i", "--isoroot",  dest="isoroot",
                      default=DEFAULT_ISOROOT,
                      help="location to store iso images (default=%s)"
                      % DEFAULT_ISOROOT)
    parser.add_option("-f", "--flavor",  dest="flavor", default=DEFAULT_FLAVOR,
                      help="flavor of Ubuntu to download (default=%s)"
                      % DEFAULT_FLAVOR)
    parser.add_option('-n', '--no-act', default=False, action='store_true',
                      dest='no_act',
                      help='compute everything but don\'t actually download')
    parser.add_option('-u', '--url', dest="base_url", default=BASE_URL,
                      help="Base URL for cdimage retrieval website (default=%s)"
                      % BASE_URL)

    (options, args) = parser.parse_args()

    if options.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # Setup URLS
    release_part = options.release
    flavor_part = options.flavor
    if options.release == DEFAULT_RELEASE:
        release_part = ''
    if options.flavor == 'ubuntu':
        flavor_part = ''
    if options.variant == 'desktop':
        l_url = os.path.join(options.base_url, release_part, flavor_part, 'daily-live')
    elif options.variant == 'alternate':
        l_url = os.path.join(options.base_url, release_part, flavor_part, 'daily')
    else:
        # Default to server
        l_url = os.path.join(options.base_url, 'ubuntu-server', release_part,
                             'daily')

    # Check options for architectures
    l_archs = []
    if options.arch == 'all':
        l_archs = ALL_ARCHS
    else:
        l_archs.append(options.arch)

    logging.info("Checking/Downloading current build for variant %s from %s",
                 options.variant, l_url)

    # Download and store sha256 digest for required ISO images
    l_current_iso_digest = {}
    sha256sum_url = os.path.join(l_url, 'current', 'SHA256SUMS')
    try:
        fh = urllib2.urlopen(sha256sum_url)
    except urllib2.HTTPError, e:
        logging.error("Failed to fetch URL '%s': %s . Aborting!",
                      sha256sum_url, e)
        sys.exit(1)

    # Process the file (will contain a number of entries
    for digest in fh:
        (chksum, sep,  iso) = digest.partition(' *')
        l_current_iso_digest[iso.rstrip('\n')] = chksum
    fh.close()

    # Check to see if version on disk is already the latest
    for arch in l_archs:
        l_iso_name = '-'.join((options.release, options.variant, arch)) + '.iso'
        l_iso_url = os.path.join(l_url, 'current', l_iso_name)

        # Check ISO availability on remote server
        for retry in range(1, RETRY_MAX+1):
            logging.debug('Try %d', retry)
            if remote_file_exists(l_iso_url):
                break

            if retry == RETRY_MAX:
                logging.info('Max number of retries reached. Aborting!')
                sys.exit(1)
            logging.info("Remote file not available. Retrying in %d minutes", TIMEOUT)
            sleep(TIMEOUT * 60)

        try:
            iso_sha256 = l_current_iso_digest[l_iso_name]
        except KeyError:
            logging.error("No digest found for image '%s'", l_iso_name)

        lock = FileLock('/tmp/' + l_iso_name)
        with lock:
            rc = 0
            l_iso_dir = options.flavor
            l_iso_location = os.path.join(options.isoroot, l_iso_dir,
                                          l_iso_name)
            logging.debug("Checking state of local cache '%s'", l_iso_location)
            # If iso does not exists then mark for retrieval
            if os.path.exists(l_iso_location):
                # Check to see if latest version
                # If digests don't match then mark for retrieval
                if not validate_sha256(l_iso_location, iso_sha256):
                    logging.debug("ISO '%s' is not the latest", l_iso_name)
                    rc = download_iso(options.isoroot, options.release,
                                      options.variant, arch, options.flavor,
                                      options.no_act)
                else:
                    logging.info("Required ISO version '%s' already in cache",
                                 l_iso_name)
            else:
                # ISO does not exist need to download
                logging.info("Downloading ISO '%s' to local cache", l_iso_name)
                rc = download_iso(options.isoroot, options.release,
                                  options.variant, arch, options.flavor,
                                  options.no_act)

            if rc != 0:
                logging.error("Failed to download ISO. Aborting!")
                sys.exit(1)
            else:
                # We downloaded something, let see if that is what we expected
                if not validate_sha256(l_iso_location, iso_sha256):
                    logging.info("Local file doesn't match checksum. Aborting!")
                    sys.exit(1)

if __name__ == '__main__':
    main()
