#/bin/sh
# Copyright (C) 2011, Canonical Ltd (http://www.canonical.com/)
#
# Automated Wubi installation testing
# 
# This software is free software: you can redistribute it 
# and/or modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation, either version 3 of 
# the License, or (at your option) any later version.
# 
# This software is distributed in the hope that it will 
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.
# 

JOBNAME="Wubi Test"

BASEDIR="$(dirname $0)/.."
BINDIR="$BASEDIR/bin"
DATADIR="$BASEDIR/data"
CUSTOMDIR="$DATADIR/custom-installation"
RUNDIR=$BASEDIR/run
LOGFILE=$RUNDIR/latest/$(basename $0 .sh).log
STATEFILE=$RUNDIR/job.state
STATE=""

. $BINDIR/functions
. $BINDIR/config

mkdir -p $( dirname $LOGFILE)

WUBIEXE="wubi.exe"

# Default Values
FLAVOR="ubuntu"
RELEASE="oneiric"
ARCH="i386"
TEST="default"
VARIANT="desktop" # Always desktop for Wubi

# Default values are overridden by those in statefile
# if a job is already running
# Then args passed on commandline
[ -f "$STATEFILE" ] && . $STATEFILE

usage() {
	cat<<EOF
Usage: $(basename $0) [OPTIONS...]

  -f, --flavor FLAVOR      Select specific flavor to test ($FLAVOR)
  -r, --release RELEASE    Select release to test ($RELEASE)
  -a, --arch ARCH          Select architecture to test ($ARCH)
  -t, --test TEST          Select test to run ($TEST)
  -h, --help               This help
EOF
	exit 1
}

run_test () {
    # Run the test
    log_info_msg "Running test for $1"
    $BINDIR/wubi-test.sh -r --isopath=$1
}

write_state() {
    STATE=$1
    cat<<EOF>${STATEFILE}
RELEASE=$RELEASE
FLAVOR=$FLAVOR
ARCH=$ARCH
TEST=$TEST
LOGFILE=$LOGFILE
STATE=$1
EOF

}

sync_results() {
    rsync -rvz --del run/latest/ $RESULTDEST
}

TEMP=$(getopt -o f:r:a:t:h --long help,flavor:,release:,arch:,test: -- "$@")
eval set -- "$TEMP"

while true ; do
    case "$1" in
        -h|--help) 
            usage;;
        -f|--flavor)
            FLAVOR=$2
            shift 2;;
        -r|--release) 
            RELEASE=$2
            shift 2 ;;
        -a|--arch) 
            ARCH=$2
            shift 2 ;;
        -t|--test)
            TEST=$2
            shift 2 ;;
        --) shift ; break ;;
        *) usage;;
    esac
done


exec 2>&1


# Log rotation on first start
if [ ! -f "$STATEFILE" ]; then
    if [ -d $RUNDIR/latest ]; then
        log_begin_msg "Backup previous log dir to $RUNDIR/$(date '+%Y%m%d-%H%M%S')"
        mv $RUNDIR/latest $RUNDIR/$(date '+%Y%m%d-%H%M%S')
        log_end_msg
    fi
    mkdir -p $RUNDIR/latest
fi

log_info_msg "Log saved to $LOGFILE"

exec 1>>$LOGFILE

log_start_job "$JOBNAME"
trap "log_end_job \"$JOBNAME\"" EXIT QUIT

JENKINS_JOB=${RELEASE}-${VARIANT}-${ARCH}_wubi-${TEST}
ISODIR=$DATADIR/iso/${FLAVOR}

FLAVORPART="${FLAVOR}/"
[  $FLAVOR == "ubuntu" ] && FLAVORPART=""

ISONAME="${RELEASE}-${VARIANT}-${ARCH}.iso"
ISOPATH=$ISODIR/$ISONAME
ZSYNCURL=http://cdimage.ubuntu.com/${FLAVORPART}daily-live/current/${ISONAME}.zsync
RESULTDEST=$JENKINS_SLAVE:/tmp/${FLAVOR}_${RELEASE}_${ARCH}_${TEST}/test-results

if [ ! -f "$STATEFILE" ]; then
    log_info_msg "State file $STATEFILE not found. Starting new job"
    # No state file, create a new job
    # Start remote jenkins task
    write_state "start-build"
    JOBURL="http://${JENKINS_URL}/job/$JENKINS_JOB/buildWithParameters?token=$JENKINS_TOKEN"
    log_info_msg "Starting remote build $JOBURL"
    ssh $JENKINS_MASTER wget "$JOBURL" -o /dev/null

    # Write state change
    # Sync result to notify start
    # Download image from cdimage.u.c
    write_state "sync"
    mkdir -p $ISODIR
    cd $ISODIR
    log_info_msg "Syncing $ZSYNCURL to $ISODIR"
    zsync $ZSYNCURL

    # Extract wubi.exe
    log_info_msg "Extracting $WUBIEXE from $ISONAME"
    bsdtar.exe xf $ISONAME $WUBIEXE
    chmod +x $WUBIEXE
    cd -

    write_state "running"

    log_info_msg "Starting Stage 1"
    # This command will restart the system, so nothing can be done after this.
    run_test $ISOPATH
elif [ "$STATE" == "running" ]; then
    log_info_msg "Starting Stage 2"
    write_state "report"
    rsync -rvz --del run/latest/ $RESULTDEST
    run_test $ISOPATH
elif [ "$STATE" == "report" ]; then
    log_info_msg "Syncing results"
    # Last step report slave and remove state file for next run
    # Sync the files
    rsync -rvz --del run/latest/ $RESULTDEST
    rm $STATEFILE
else
    log_warning_msg "Unknown state: $STATE"
    log_warning_msg "Aborting!"
    exit 1
fi

log_end_msg 
