#/bin/sh
# Copyright (C) 2011, Canonical Ltd (http://www.canonical.com/)
#
# Automated Wubi installation testing
# 
# This software is free software: you can redistribute it 
# and/or modify it under the terms of the GNU General Public License 
# as published by the Free Software Foundation, either version 3 of 
# the License, or (at your option) any later version.
# 
# This software is distributed in the hope that it will 
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this software.  If not, see <http://www.gnu.org/licenses/>.
# 

# TODO: 
#  - Verify that Windows stage is complete and write results somewhere (Use unittest ?)
#    - The following files are on C:/ : wubildr and wubildr.mbr
#    - TARGET_DRIVE/ubuntu is a directory
#    - Verify the content of $TARGET_DRIVE/ubuntu
#  - Add other releases


JOBNAME="Wubi Test"

BASEDIR="$(dirname $0)/.."
BINDIR="$BASEDIR/bin"
DATADIR="$BASEDIR/data"
CUSTOMDIR="$DATADIR/custom-installation"
RUNDIR=$BASEDIR/run
LOGFILE=$RUNDIR/latest/$(basename $0 .sh).log

. $BINDIR/functions

DRIVE="c"
I386=""
ISOPATH=""
DIMAGEPATH=""
REBOOT=""
USERNAME="ubuntu"
PASSWORD="ubuntu"
WUBIEXE="$DATADIR/iso/ubuntu/wubi.exe"

TIMEOUT=3600 # Wubi run asynchronously. Need to wait to it to finish and download takes time

usage() {
	cat<<EOF
Usage: $(basename $0) [OPTIONS...]

  -d, --dimagepath=FILE    Use specified diskimage (Downloaded if unspecified)
  -i, --isopath=FILE       Use specified ISO (Downloaded if unspecified)
  -r, --reboot             Reboot after installation
  -t, --drive=TARGET_DRIVE Target drive (Default ${DRIVE}:)
  -x, --32bit              Force installation of 32bit version
  -h, --help               This help
EOF
	exit 1
}

get_bootid() {
    # Return Window's bootid for the boot entry with 'name'

    name=$1
    if [ -z "$name" ]; then
        return
    fi
    bootid=$( $BINDIR/bcdedit.exe |grep -B3 "$name" |grep identifier|awk '{print $2}' )
    echo $bootid
}

TEMP=$(getopt -o hd:i:rt:x --long help,dimagepath:,drive:,reboot,isopath:,32bit -- "$@")
eval set -- "$TEMP"

while true ; do
    case "$1" in
        -h|--help) 
            usage;;
        -t|--drive) 
            DRIVE=$2
            shift 2 ;;
        -d|--dimagepath) 
            DIMAGEPATH=$2
            shift 2 ;;
        -i|--isopath) 
            ISOPATH=$2
            shift 2 ;;
        -r|--reboot)
            REBOOT="Y"
            shift;;
        -x|--32bit)
            I386="--32bit"
            shift;;
        --) shift ; break ;;
        *) usage;;
    esac
done

#exec 2>&1

if [ -f "$RUNDIR/latest/stage2" ]; then
    #exec 1>>$LOGFILE
    log_start_job "$JOBNAME : Stage 2"

    # Just grep 'buntu' to catch derivatives as well
    # It will be used to set default boot entry for stage 2 of the test
    bootid=$(get_bootid "buntu")

    if [ -z "$bootid" ]; then
        log_failure_msg "Boot entry for Ubuntu not found. Aborting."
        exit 3
    fi

    log_info_msg "Setting $bootid entry as default for next reboot"
    $BINDIR/bcdedit.exe /bootsequence $bootid

    if [ "$REBOOT" == "Y" ]; then
        # Reboot
        log_info_msg "Rebooting"
        shutdown.exe /r /f /t 5 
    fi
    log_end_msg 
    exit 0
fi

#if [ -d $RUNDIR/latest ]; then
#    log_begin_msg "Backup previous log dir to $RUNDIR/$(date '+%Y%m%d-%H%M%S')"
#    mv $RUNDIR/latest $RUNDIR/$(date '+%Y%m%d-%H%M%S')
#    log_end_msg
#fi
#mkdir -p $RUNDIR/latest

#log_info_msg "Log saved to $LOGFILE"

#exec 1>>$LOGFILE

log_start_job "$JOBNAME"
trap "log_end_job \"$JOBNAME\"" EXIT QUIT

if [ ! -x "$WUBIEXE" ]; then
    log_failure_msg "'$WUBIEXE' not found"
    exit 1
fi

#WUBIARGS="--password=$PASSWORD --username=$USERNAME --noninteractive --skipmd5check --log-file=$WUBILOG"
WUBIARGS="--password=$PASSWORD --username=$USERNAME --noninteractive --skipmd5check"
if [ ! -z "$DIMAGEPATH" ]; then
    if [ ! -f "$DIMAGEPATH" ]; then
        log_failure_msg "File doesn't exist '$DIMAGEPATH'"
        exit 1
    fi

    # If an image and 32bit are specified on the command line then check that the image is not 64bit
    if basename "$DIMAGEPATH" | grep -q "amd64"; then
        if [ ! -z "$I386" ]; then
            log_failure_msg "You forced installation of 32bit version but the image is 64bit" 
            exit 1
        fi
    else
        I386="--32bit"
    fi
elif [ ! -z "$ISOPATH" ]; then
    if [ ! -f "$ISOPATH" ]; then
        log_failure_msg "File doesn't exist '$ISOPATH'"
        exit 1
    fi

    MIMETYPE=$( file -bi "$ISOPATH" )
    if ! echo "$MIMETYPE" | grep -q iso9660; then
        log_failure_msg "'$ISOPATH' is not a valid ISO (mimetype: $MIMETYPE)"
        exit 1
    fi

    # If an image and 32bit are specified on the command line then check that the image is not 64bit
    if basename "$ISOPATH" | grep -q "amd64"; then
        if [ ! -z "$I386" ]; then
            log_failure_msg "You forced installation of 32bit version but the image is 64bit" 
            exit 1
        fi
    else
        I386="--32bit"
    fi
fi

if [ ! -z "$DIMAGEPATH" ]; then
    DIMAGEPATH="--dimagepath=$(cygpath -aw $DIMAGEPATH)"
    WUBIARGS="$WUBIARGS $I386 $DIMAGEPATH"
elif [ ! -z "$ISOPATH" ]; then
    ISOPATH="--isopath=$(cygpath -aw $ISOPATH)"
    WUBIARGS="$WUBIARGS $I386 $ISOPATH"
fi


log_begin_msg "Uninstalling previous install"
$WUBIEXE --uninstall --noninteractive

# Take some breath
sleep 5
log_end_msg

log_info_msg "Running $(basename $WUBIEXE) $WUBIARGS"
$WUBIEXE $WUBIARGS

STARTTIME=$(date +%s)
while true; do
    currtime=$(date +%s)
    timediff=$(( $currtime - $STARTTIME ))

    if [ $timediff -gt $TIMEOUT ]; then
        log_failure_msg "Wubi still running after $TIMEOUT seconds. Aborting!"
        exit 2
    fi
    if ! ps -W | grep -q "pyrun.exe"; then
        log_info_msg "Wubi installation done (took ${timediff}s)"
        break
    fi
    log_info_msg "Waiting for wubi to finish. running for ${timediff}s"
    sleep 1
done

# Only ISO based installation (which uses casper) supports the
# custom-installation facility
if [ ! -z "$ISOPATH" ]; then
    TARGET_CUSTOMDIR=/cygdrive/${DRIVE}/ubuntu/install/custom-installation

    # Load casper custome-installation directory with tests
    log_begin_msg "Copying $(basename $CUSTOMDIR) to $TARGET_CUSTOMDIR"
    cp -af $CUSTOMDIR/* $TARGET_CUSTOMDIR || true
    log_end_msg
fi

log_begin_msg "Saving Wubi log file"
mv /tmp/wubi*.log $RUNDIR/latest/
log_end_msg

log_begin_msg "Preparing for stage 2"
touch $RUNDIR/latest/stage2
log_end_msg

if [ "$REBOOT" == "Y" ]; then
    # Reboot
    log_info_msg "Rebooting"
    shutdown.exe /r /f /t 5 
fi

log_end_msg 
